{*******************************************************}
{                                                       }
{       AllPer - Simplo Components Library              }
{       Simple XML Access Component Unit                }
{                                                       }
{       Copyright (c) 2003, AllPer Software Engineering }
{                                                       }
{       Author: Eduardo Alves                           }
{       e-allper@uol.com.br
{                                                       }
{*******************************************************}
unit SimploXML;

interface
uses classes, Sysutils, cStrings, SimploXMLUtils;

type
  TElementDocXML = class;
  TBlockDocXML = array of TElementDocXML;

  TElementDocXML = class(TObject)
  private
    FText: string;
    FName: string;
    function GetFirstElement: TElementDocXML;
  protected
    FElements: TBlockDocXML;
  public

    property Elements: TBlockDocXML read FElements;
    property Name: string read FName write FName;
    property Text: string read FText write FText;

    function AddElement(ElementName:string): TElementDocXML;
    function ElementsByName(ElementName:string): TBlockDocXML;

  end;

  TSimploXML = class(TComponent)
  private
    FDoc: TElementDocXML;
    FOnXMLError: TNotifyEvent;
    function GetRoot: TElementDocXML;

  public
    property Root: TElementDocXML read GetRoot;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function AsString: string;
    function LoadFromFile(const FileName: string): boolean;
    function LoadFromString(const XMLText: string): boolean;

    procedure Clear;
    procedure CreateRoot(const RootName: string);

  published
    property OnXMLError: TNotifyEvent read FOnXMLError write FOnXMLError;

  end;

procedure Register;

implementation

{ TSimploXML }

function TSimploXML.AsString: string;
  procedure WriteElement(CurrentElement: TElementDocXML);
  var
    i: integer;
  begin
    Result := Result+XMLQuoted(CurrentElement.Name);
    Result := Result+CurrentElement.Text;
    for i := 0 to length(CurrentElement.Elements)-1 do
    begin
      WriteElement(CurrentElement.Elements[i]);
    end;
    Result := Result+XMLQuoted('/'+CurrentElement.Name);
  end;

begin
  Result := '<?xml version="1.0" encoding="ISO-8859-1"?>';
  WriteElement(GetRoot);
end;

procedure TSimploXML.Clear;
var
  i: integer;
begin
  for i := length(FDoc.FElements)-1 downto 0  do
    FDoc.FElements[i].Free;
  SetLength(FDoc.FElements, 0);
end;

constructor TSimploXML.Create;
begin
  inherited;
  FDoc := TElementDocXML.Create;
end;

procedure TSimploXML.CreateRoot(const RootName: string);
begin
  Clear;
  FDoc.AddElement(RootName);
end;

destructor TSimploXML.Destroy;
begin
  FDoc.free;
  inherited;
end;

function TSimploXML.GetRoot: TElementDocXML;
begin
  Result := FDoc.GetFirstElement;
end;

function TSimploXML.LoadFromFile(const FileName: string): boolean;
var
  FileStream: TFileStream;
  DocText: string;

begin
  if FileExists(FileName) then
  begin
    FileStream := TFileStream.Create(FileName, fmOpenRead);
    try
      try
        SetLength(DocText, FileStream.Size);
        FileStream.ReadBuffer(Pointer(DocText)^, FileStream.Size);
      finally
        FileStream.free;
      end;
      Result := LoadFromString(DocText);
    except
      Clear;
      Result := false;
      if assigned(FOnXMLError) then
        FOnXMLError(self);
    end;
  end
  else
  begin
    Clear;
    Result := false;
    if assigned(FOnXMLError) then
      FOnXMLError(self);
  end;
end;

function TSimploXML.LoadFromString(const XMLText: string): boolean;
var
  DocText: string;
  TagList: TTagList;
  i, RootTagIndex, CurrentTagIndex: integer;

  procedure UpdateElement(CurrentElement: TElementDocXML);
  var
    OldElement: TElementDocXML;
  begin
    if (TagList[CurrentTagIndex].Text <> '/'+CurrentElement.Name) then
    begin
      if TagList[CurrentTagIndex].IsTag then
      begin
        OldElement := CurrentElement;
        CurrentElement := CurrentElement.AddElement(TagList[CurrentTagIndex].Text);
        Inc(CurrentTagIndex);
        UpdateElement(CurrentElement);
        if OldElement.Name <> '' then // FDoc
          UpdateElement(OldElement);
      end
      else
      begin
        CurrentElement.Text := TagList[CurrentTagIndex].Text;
        Inc(CurrentTagIndex);
        UpdateElement(CurrentElement);
      end;
    end
    else
      Inc(CurrentTagIndex);
  end;

begin
  try
    if XMLText = '' then
      Abort;

    Clear;
    DocText := XMLText;
    ExtractXMLTags(TagList, DocText);

    // Search the root Element
    RootTagIndex := -1;
    for i := 0 to length(TagList)-1 do
    begin
      if TagList[i].IsTag then
      begin
        RootTagIndex := i;
        Break;
      end;
    end;

    if (RootTagIndex >= 0) then
    begin
      CurrentTagIndex := RootTagIndex;
      UpdateElement(FDoc);
    end;
    Result := true;
  except
    Clear;
    Result := false;
    if assigned(FOnXMLError) then
      FOnXMLError(self);
  end;
end;

{ TElementDocXML }

function TElementDocXML.AddElement(ElementName:string): TElementDocXML;
begin
  SetLength(FElements, length(FElements)+1);
  FELements[length(FElements)-1] := TElementDocXML.Create;
  Result := FELements[length(FElements)-1];
  Result.Name := ElementName;
end;

function TElementDocXML.ElementsByName(ElementName: string): TBlockDocXML;
var
  i: integer;
begin
  setlength(Result, 0);
  for i := 0 to length(FElements)-1 do
  begin
    if FElements[i].Name = ElementName then
    begin
      setlength(Result, length(Result)+1);
      Result[length(Result)-1] := FElements[i];
    end;
  end;
end;

function TElementDocXML.GetFirstElement: TElementDocXML;
begin
  if assigned(FElements) then
    Result := FElements[0]
  else
    Result := nil;
end;

procedure Register;
begin
  RegisterComponents('Simplo', [TSimploXML]);
end;

end.
