unit XLLists;

interface uses Classes;
                                  
type
  PPointer = ^Pointer;

{ Apply function for the applicator:
     TheIndex        Index of item in array
     TheItem         Value of item (i.e pointer element) in section
     Returns: 0 if success, else error code. }

  TSPAApply = function(TheIndex: Integer; TheItem: Pointer): Integer;

  PSecDir = ^TSecDir;
  TSecDir = array[0..4095] of Pointer;  { Enough for up to 12 bits of sec }

{ TSList class - array of pointers}

  TSList = class
  private
    secDir: PSecDir;
    slotsInDir: Word;
    FSectionSize: Word;
    indexMask, secShift: Word;
    FHighBound: Integer;
    cachedIndex: Integer;
    cachedPointer: Pointer;
    FCount: Integer;
    FLarge: Boolean;
    procedure DoneData;
    procedure InitData;
    function  GetAt(Index: Integer): TObject;
    function  MakeAt(Index: Integer): PPointer;
    procedure PutAt(Index: Integer; Item: TObject);
    procedure SetCount(ACount: Integer);
  public
    constructor Create;
    destructor  Destroy; override;
    procedure DestroyItem(AItem: TObject); virtual;
    function  ForAll(ApplyFunction: Pointer): Integer;
    procedure StreamLoad(AStream: TStream); virtual; // read pointers as integers
    procedure StreamSave(AStream: TStream); virtual; // write pointers as integers
    procedure Modify(AIndex, ACount: Integer); // insert or remove Abs(ACount) items from AIndex
    procedure ResetHighBound;
    function  Empty: Boolean;
    property  HighBound: Integer read FHighBound;
    property  SectionSize: Word read FSectionSize;
    property  Count: Integer read FCount write SetCount;
    property  Items[Index: Integer]: TObject read GetAt write PutAt; default;
    procedure Renumerate(AFrom, ATo: Integer); virtual;
    procedure Clear; virtual;
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    procedure Insert(Index: Integer; Item: TObject);
    procedure Move(CurIndex, NewIndex: Integer);
    procedure MovePair(CurIndex, NewIndex: Integer);
  end;

  TSListClass = class of TSList;

{ TZeroSList class - array of Integers with Zero offset}
  TProZeroSList = class(TSList)
  private
    FZero: Integer;
    procedure SetZero(AZero: Integer); virtual;
  public
    property  Zero: Integer read FZero write SetZero;
  end;

  TZeroSList = class(TProZeroSList)
  private
    function  GetInt(AIndex: Integer): Integer;
    procedure SetInt(AIndex, AValue: Integer);
    procedure SetZero(AZero: Integer); override;
  public
    function  SummTo(ATo: Integer): Integer;
    function  ScanIndex(var ASumm: Integer): Integer;
    property  Ints[AIndex: Integer]: Integer read GetInt write SetInt; default;
  end;


{ TDefSList class - array of Integers with Default value}

  TDefSList = class(TProZeroSList)
  private
    function  GetInt(AIndex: Integer): Integer;
    procedure SetInt(AIndex, AValue: Integer);
  public
    property  Defs[AIndex: Integer]: Integer read GetInt write SetInt; default;
  end;

{ TObjectSList - array of TObject}
  TObjectSList = class(TSList) // Items will be destroyed
  public
    procedure DestroyItem(AItem: TObject); override;
    procedure Clear; override;
  end;

{ TListSList - array of TSList}
  TListSList = class(TObjectSList)
  private
    FClass: TSListClass;
  public
    procedure Prepare(ASList: TSList; AIndex: Integer); virtual;
    function Ensure(AIndex: Integer): TSList;
    constructor Create(AClass: TSListClass);
    procedure StreamLoad(AStream: TStream); override; // read TSLists
    procedure StreamSave(AStream: TStream); override; // write TSLists
    property ListClass: TSListClass read FClass;
  end;

  function  IMax(A,B: Integer): Integer;
  function  IMin(A,B: Integer): Integer;
  procedure FreeObject(var Obj);

const
   LargeArrays: Boolean = False;

implementation

function IMin(A, B: Integer): Integer; register;
begin
  Result := B;
  if A < B then Result := A;
end;

function IMax(A, B: Integer): Integer; register;
begin
  Result := B;
  if A > B then Result := A;
end;

procedure FreeObject(var Obj);
begin
  TObject(Obj).Free;
  Pointer(Obj) := Nil;
end;

const
  SPAIndexMask: array[Boolean] of Byte = (15, 255);
  SPASecShift: array[Boolean] of Byte = (4, 8);

function  ExpandDir(secDir: PSecDir; var SlotsInDir: Word; NewSlots: Word): PSecDir;
begin
  Result := SecDir;
  ReallocMem(Result, NewSlots * SizeOf(Pointer));
  FillChar(Result^[SlotsInDir], (NewSlots - SlotsInDir) * SizeOf(Pointer), 0);
  SlotsInDir := NewSlots;
end;

function  MakeSec(SecIndex: Integer; SectionSize: Word): Pointer;
var
  SecP: Pointer;
  Size: Word;
begin
  Size := SectionSize * SizeOf(Pointer);
  GetMem(SecP, Size);
  FillChar(SecP^, Size, 0);
  MakeSec := SecP;
end;

procedure   TSList.InitData;
begin
  FSectionSize := Word(SPAIndexMask[FLarge]) + 1;
  IndexMask := Word(SPAIndexMask[FLarge]);
  SecShift := Word(SPASecShift[FLarge]);
  SecDir := nil;
  SlotsInDir := 0;
  FHighBound := -1;
  CachedIndex := -1;
end;

constructor TSList.Create;
begin
  FLarge := LargeArrays;
  InitData;
end;

procedure TSList.DestroyItem(AItem: TObject);
begin
end;

procedure  TSList.DoneData;
var
  I:  Integer;
  Size: Word;
begin
  I := 0;
  Size := FSectionSize * SizeOf(Pointer);
  while I < SlotsInDir do
  begin
    if SecDir^[I] <> nil then FreeMem(SecDir^[i], Size);
    Inc(I);
  end;
  if SecDir <> nil then FreeMem(SecDir, SlotsInDir * SizeOf(Pointer));
end;

function  TSList.GetAt(Index: Integer): TObject;
var
  ByteP: PChar;
  SecIndex: Cardinal;
begin
  Result := Nil;
  if (Index < 0) or (Index > HighBound) then exit;
  if Index = CachedIndex then Result := CachedPointer else
  begin
    SecIndex := Index shr SecShift;
    if SecIndex >= SlotsInDir then ByteP := Nil else
    begin
      ByteP := SecDir^[SecIndex];
      if ByteP <> nil then Inc(ByteP, (Index and IndexMask) * SizeOf(Pointer));
    end;
    if ByteP <> nil then Result := PPointer(byteP)^;
    CachedIndex := Index;
    CachedPointer := Result;
  end
end;

function  TSList.MakeAt(Index: Integer): PPointer;
var
  DirP: PSecDir;
  P: Pointer;
  ByteP: PChar;
  SecIndex: Word;
begin
  SecIndex := Index shr SecShift;
  if SecIndex >= SlotsInDir
  then DirP := ExpandDir(SecDir, SlotsInDir, SecIndex + 1)
  else DirP := SecDir;
  { Index into Section Directory using high order part of
    index.  Get pointer to Section. If null, create new
    Section.  Index into Section using low order part of index. }
  SecDir := DirP;
  P := DirP^[SecIndex];
  if P = nil then
  begin
    P := MakeSec(SecIndex, FSectionSize);
    DirP^[SecIndex] := P;
  end;
  ByteP := P;
  Inc(ByteP, (Index and IndexMask) * SizeOf(Pointer));
  if Index > FHighBound then FHighBound := Index;
  Result := PPointer(ByteP);
  CachedIndex := -1
end;

procedure TSList.PutAt(Index: Integer; Item: TObject);
begin
  if (Index >= 0) and ((Item <> nil) or (GetAt(Index) <> nil)) then
  begin
    MakeAt(Index)^ := Item;
    if Item = nil then ResetHighBound;
  end;
end;

function  TSList.ForAll(ApplyFunction: Pointer {TSPAApply}): Integer;
var
  ItemP: PChar;
  Item: Pointer;
  I, CallerBP: Cardinal;
  J, Index: Integer;
begin
  Result := 0;
  I := 0;
  asm
    mov   eax,[ebp]                     { Set up stack frame for local }
    mov   callerBP,eax
  end;
  while (I < SlotsInDir) and (Result = 0) do
  begin
    ItemP := SecDir^[i];
    if ItemP <> nil then
    begin
      J := 0;
      Index := I shl SecShift;
      while (J < FSectionSize) and (Result = 0) do
      begin
        Item := PPointer(ItemP)^;
        if Item <> nil  then
          { ret := ApplyFunction(index, item.Ptr); }
          asm
            mov   eax,Index
            mov   edx,Item
            push  CallerBP
            call  ApplyFunction
            pop   ecx
            mov   @Result,eax
          end;
        Inc(ItemP, SizeOf(Pointer));
        Inc(J);
        Inc(Index)
      end;
    end;
    Inc(I);
  end;
end;

procedure TSList.ResetHighBound;
var
  NewHighBound: Integer;

  function  Detector(TheIndex: Integer; TheItem: Pointer): Integer; far;
  begin
    if TheIndex > FHighBound
    then Result := 1
    else begin
           Result := 0;
           if TheItem <> nil
           then NewHighBound := TheIndex;
         end;
  end;

begin
  NewHighBound := -1;
  ForAll(@Detector);
  FHighBound := NewHighBound;
end;

destructor TSList.Destroy;
begin
  Clear;
end;

procedure TSList.Clear;

begin
  DoneData;
  InitData;
end;

procedure TSList.Delete(Index: Integer);
var
  I, Max: Integer;
begin
  if (Index < 0) or (Index > HighBound) then Exit;
  Max := HighBound;
  for I := Index to Max do Items[I] := Items[I + 1];
  Renumerate(Index, Max);
end;

procedure TSList.Exchange(Index1, Index2: Integer);
var
  temp: TObject;
begin
  temp := GetAt(Index1);
  PutAt(Index1, GetAt(Index2));
  PutAt(Index2, temp);
  Renumerate(Index1, Index1);
  Renumerate(Index2, Index2);
end;

function  TSList.Empty: Boolean;
begin
  Result := (HighBound < 0);
end;

procedure TSList.Insert(Index: Integer; Item: TObject);
var
  I: Integer;
begin
  if Index < 0 then exit;
  I := HighBound + 1;
  while I > Index do
  begin
    Items[I] := Items[I - 1];
    Dec(I);
  end;
  Items[Index] := Item;
  Renumerate(Index + 1, FHighBound);
  Renumerate(Index, Index);
end;

procedure TSList.MovePair(CurIndex, NewIndex: Integer);
var Item1,
    Item2,
    ItemK: TObject;
begin
  if CurIndex <> NewIndex then
  begin
    Item1 := GetAt(CurIndex);
    Item2 := GetAt(CurIndex + 1);
    Delete(CurIndex);
    PutAt(CurIndex, Nil);
    ItemK := GetAt(NewIndex);
    PutAt(NewIndex, Item2);
    Insert(NewIndex, Item1);
    ItemK.Free;
  end;
end;

procedure TSList.Move(CurIndex, NewIndex: Integer);
var
  Item: Pointer;
begin
  if CurIndex <> NewIndex then
  begin
    Item := GetAt(CurIndex);
    Delete(CurIndex);
    Insert(NewIndex, Item);
  end;
end;

procedure TSList.Renumerate(AFrom, ATo: Integer);
begin
end;

procedure TSList.StreamLoad(AStream: TStream);  // read integers
var
  SlotsToRead,
  I, J: Integer;
  ItemP: Pointer;
begin
  Clear;
  AStream.Read(FLarge, SizeOf(FLarge));
  AStream.Read(FCount, SizeOf(FCount));
  InitData;
  AStream.Read(SlotsToRead, SizeOf(Integer));
  for I := 1 to SlotsToRead do
  begin
    AStream.Read(J, SizeOf(Integer));
    MakeAt(J shl SecShift);
    ItemP := SecDir^[J];
    AStream.Read(ItemP^, FSectionSize * SizeOf(Pointer));
    FHighBound := (J + 1) shl SecShift;
  end;
  ResetHighBound;
end;

procedure TSList.StreamSave(AStream: TStream);    // write integers
var
  SlotsToSave, I: Integer;
  ItemP: Pointer;
begin
  AStream.Write(FLarge, SizeOf(FLarge));
  AStream.Write(FCount, SizeOf(FCount));
  SlotsToSave := 0;
  for I := 0 to SlotsInDir - 1 do
  if SecDir^[I] <> Nil then Inc(SlotsToSave);
  AStream.Write(SlotsToSave, SizeOf(Integer));
  for I := 0 to SlotsInDir - 1 do
  begin
    ItemP := SecDir^[I];
    if ItemP <> Nil then
    begin
      AStream.Write(I, SizeOf(I));
      AStream.Write(ItemP^, FSectionSize * SizeOf(Pointer));
    end;
  end;
end;

procedure TSList.SetCount(ACount: Integer);
begin
  FCount := ACount;
  while HighBound >= ACount do
  begin
    DestroyItem(Items[HighBound]);
    Items[HighBound] := Nil;
  end;
end;

procedure TSList.Modify(AIndex, ACount: Integer);
var I: Integer;
begin
  for I :=  1 to ACount do Insert(AIndex, Nil);
  for I := -1 downto ACount do begin DestroyItem(Items[AIndex]); Delete(AIndex); end;
end;

// TProZeroSList
procedure TProZeroSList.SetZero(AZero: Integer);
begin
  FZero := AZero;
end;

// TZeroSList

function  TZeroSList.GetInt(AIndex: Integer): Integer;
begin
  Result := Integer(GetAt(AIndex)) + FZero;
end;

procedure TZeroSList.SetInt(AIndex, AValue: Integer);
begin
  PutAt(AIndex, Pointer(AValue - FZero));
end;

function  TZeroSList.SummTo(ATo: Integer): Integer;
var Idx: Integer;
    ItemP: PChar;
    I, J: Integer;
begin
  Result := 0;
  if ATo > 0 then
  begin
    Result := ATo * FZero;
    Idx := 0;
    for I := 0 to SlotsInDir - 1 do
    begin
      ItemP := SecDir^[I];
      // All are zero?
      if ItemP = nil then
      begin
        Inc(Idx, FSectionSize);
        if Idx > ATo then exit;
      end else
      begin
        // Summ all section
        if Idx + FSectionSize < ATo then
        begin
          for J := 0 to FSectionSize - 1 do
          begin
            Inc(Result, Integer(PPointer(ItemP)^));
            Inc(ItemP, SizeOf(Pointer));
          end;
          Inc(Idx, FSectionSize);
        end else
        begin
          for J := 1 to ATo - Idx do
          begin
            Inc(Result, Integer(PPointer(ItemP)^));
            Inc(ItemP, SizeOf(Pointer));
          end;
          exit;
        end;
      end;
    end;
  end;
end;

function  TZeroSList.ScanIndex(var ASumm: Integer): Integer;
var Summ: Integer;
    ItemP: PChar;
    I,J,M: Integer;
    SSZ: Integer;
begin
  if (ASumm > 0) and (FZero > 0) then
  begin
    SSZ := FSectionSize * FZero;
    Summ := ASumm;
    Result := 0;
    I := 0;
    while True do
    begin
      if I < SlotsInDir then ItemP := SecDir^[I] else ItemP := Nil;
      // All are zero?
      if ItemP = nil then
      begin
        if Summ >= SSZ then
        begin
          Dec(Summ, FSectionSize * FZero);
          Inc(Result, FSectionSize);
        end else
        begin
          {Somewhere in range}
          Inc(Result, Summ div FZero);
          Dec(ASumm, Summ mod FZero);
          exit;
        end;
      end else
      begin
      // All present
        for J := 0 to FSectionSize - 1 do
        begin
          M := FZero + Integer(PPointer(ItemP)^);
          Dec(Summ, M);
          if Summ < 0 then
          begin
            {Found}
            Inc(Result, J);
            Dec(ASumm, M + Summ);
            exit;
          end;
          Inc(ItemP, SizeOf(Pointer));
        end;
        Inc(Result, FSectionSize);
      end;
      Inc(I);
    end;
  end;
  ASumm := 0;
  Result := 0;
end;

procedure TZeroSList.SetZero(AZero: Integer);
begin
  inherited SetZero(AZero);
  Clear;
end;

// TDefSList

function  TDefSList.GetInt(AIndex: Integer): Integer;
begin
  Result := Integer(GetAt(AIndex));
  if Result = 0 then Result := FZero else Dec(Result);
end;

procedure TDefSList.SetInt(AIndex, AValue: Integer);
begin
  if AValue = FZero then AValue := 0 else Inc(AValue);
  PutAt(AIndex, Pointer(AValue));
end;

// TObjectList

procedure TObjectSList.Clear;

  function  FreeItem(TheIndex: Integer; TheItem: TObject): Integer;
  begin
    TheItem.Free;
    Result := 0;
  end;

begin
  ForAll(@FreeItem);
  inherited Clear;
end;

procedure TObjectSList.DestroyItem(AItem: TObject);
begin
  AItem.Free;
end;

// TListSList

constructor TListSList.Create(AClass: TSListClass);
begin
  inherited Create;
  FClass := AClass;
end;

function  TListSList.Ensure(AIndex: Integer): TSList;
begin
  Result := TSList(GetAt(AIndex));
  if Result = Nil then begin
    Result := ListClass.Create;
    PutAt(AIndex, Result);
    Prepare(Result, AIndex);
  end;
end;

procedure TListSList.Prepare(ASList: TSList; AIndex: Integer);
begin
  {Do nothing}
end;

procedure TListSList.StreamLoad(AStream: TStream);  // read TSLists
var ListsToRead, I, J: Integer;
    NL: TSList;
begin
  AStream.Read(ListsToRead, SizeOf(Integer));
  for I := 1 to ListsToRead do
  begin
    AStream.Read(J, SizeOf(Integer));
    NL := Ensure(J);
    NL.StreamLoad(AStream);
  end;
end;

procedure TListSList.StreamSave(AStream: TStream);    // write TSLists
var ListsToSave: Integer;

  function  Calculate(TheIndex: Integer; TheItem: TSList): Integer; far;
  begin
    if TheItem.HighBound >= 0 then Inc(ListsToSave);
    Result := 0;
  end;

  function  SaveList(TheIndex: Integer; TheList: TSList): Integer; far;
  begin
    if TheList.HighBound >= 0 then
    begin
      AStream.Write(TheIndex, SizeOf(Integer));
      TheList.StreamSave(AStream);
    end;
    Result := 0;
  end;

begin
  ListsToSave := 0;
  ForAll(@Calculate);
  AStream.Write(ListsToSave, SizeOf(Integer));
  ForAll(@SaveList);
end;

end.
