unit FileMapping;
// This unit implements Windows file mapping. File mapping is supported on
// regular files and using the page file. Sharing of file mapping objects is
// supported as well. The maximum size of file mapping objects is virtually
// unlimited (16 exa bytes to be exact, try to cram that on a single logical
// device). Views on the map are limited in size by the largest block of
// contiguous virtual address space, which on NT is 2Gb minus what's already
// being used and 1Gb on W9x minus what's already in use.
//
// The code in this unit is copy-lefted to you by Patrick Van Laake under the
// GPL license (http://www.gnu.org/copyleft/gpl.html). In plain English this
// means that you can freely use any of the code supplied to you, including
// modifying it and further distributing it, as long as Patrick Van Laake is
// acknowledged as the original author and no restrictions beyond any in the GPL
// are placed on it or on any products derived from it. If you wish to use this
// code in a commercial product, please contact Patrick Van Laake directly.
//
// Author:  Patrick van Laake
//          patrick.vanlaake@ieee.org
//
// Version: NOTE--VERSION 0 MEANS DEVELOPMENT CODE
//          0.2  13Jan03  Corrected logical error in TFileMapping.Create
//                        Removed tests on parameter sanity, let the OS worry
//                        Enter FSizeHi, unlimited file size support
//                        Security attributes are now passed through to the OS
//          0.1  10Jan03  Initial coding, tested, not stressed (mutexes)

interface
uses
  Windows, SysUtils, Classes;

type
  EFileMapping = class(Exception);

  TMutex = class
  private
    FHandle   : THandle;
    FTimeOut  : cardinal;
  public
    constructor Create(const name: string);
    destructor  Destroy; override;

    function    Get: boolean;
    function    Release: boolean;
    property    TimeOut: cardinal read FTimeOut write FTimeOut;
  end;

  TFileMapping = class;

  TFileView = class
  private
    FFileMap  : TFileMapping;
    FMemory   : pointer;
    FCurrent  : pointer;
    FPosition : cardinal;
    FMutex    : TMutex;
    FSize     : cardinal;
  private
    procedure   SetPosition(pos: cardinal);
  public
    constructor Create(map: TFileMapping; mtx: TMutex; access, size, start: cardinal; starthi: cardinal = 0);
    destructor  Destroy; override;

    function    Read(buffer: pointer; count: cardinal): cardinal;
    function    Write(buffer: pointer; count: cardinal): cardinal;
    function    WriteSafe(buffer: pointer; count: cardinal): cardinal;
    function    Flush: boolean;

    function    Lock: boolean;
    procedure   Unlock;

    property    BaseAddress: pointer read FMemory;
    property    Position: cardinal read FPosition write SetPosition;
  end;

  TFileMapping = class
  private
    FFileHandle : THandle;
    FMapHandle  : THandle;
    FSize       : cardinal;
    FHiSize     : cardinal;
    FPageSize   : cardinal;
    FProtect    : cardinal;
    FShared     : boolean;
  public
    constructor Create(const fn, name: string; sec: PSecurityAttributes;
                       protect: cardinal; size: cardinal; hisize: cardinal = 0);
    destructor  Destroy; override;

    function    MakeView(mtx: TMutex; start: cardinal; starthi: cardinal = 0;
                         size: cardinal = 0; access: cardinal = 0): TFileView;
  end;

implementation
///////////////////////////////////////////////////////////////////////////////
// TMutex
constructor TMutex.Create(const name: string);
begin
  FHandle := CreateMutex(nil, false, PChar(name));
  if FHandle = 0 then
    raise EFileMapping.CreateFmt('Could not create mutex %s', [name]);

  FTimeOut := 1000; // 1 second default time out
end;

destructor TMutex.Destroy;
begin
  Release;
  CloseHandle(FHandle);
end;

function TMutex.Get: boolean;
begin
  Result := WaitForSingleObject(FHandle, FTimeOut) = WAIT_OBJECT_0;
end;

function TMutex.Release: boolean;
begin
  Result := ReleaseMutex(FHandle);
end;

//////////////////////////////////////////////////////////////////////////////
// TFileMapping
constructor TFileMapping.Create(const fn, name: string; sec: PSecurityAttributes;
                                protect, size: cardinal; hisize: cardinal = 0);
// Create a file mapping object from a file name or on the page file. The file
// will be opened with access rights compatible with the protect parameter.
// Share mode will be set to 0 for exclusive access. Section bits are not
// checked but passed on the Windows call. If the name parameter is given shared
// use of the views to the mapped object is assumed and a mutex is required for
// reading to and writing from any views based on this file mapping object.
var access : cardinal;
    sysInfo : TSystemInfo;
begin
  FFileHandle := INVALID_HANDLE_VALUE;
  GetSystemInfo(sysInfo);
  FPageSize := sysInfo.dwPageSize;
  FProtect := protect;
  FShared := name <> '';

  // If shared access is requested, check if the named object already exists
  if FShared then
    FMapHandle := OpenFileMapping(8 div (FProtect and $E), false, PChar(name));
  if FMapHandle > 0 then begin
    FSize := size;     // !!!!!!!!!!!!! size must be accurate value, not 0
    FHiSize := hisize;
    Exit;
  end;

  // Open the disk file or use the page file
  if fn = '' then begin  // use the page file
    if size = 0 then
      raise EFileMapping.Create('Invalid size for page file file mapping object');
    FSize := size;
    FHiSize := 0;        // page file use is size limited
  end else begin         // use disk file
    if protect and (PAGE_READWRITE or PAGE_WRITECOPY) > 0 then
      access := GENERIC_READ or GENERIC_WRITE
    else if protect and PAGE_READONLY = PAGE_READONLY then
      access := GENERIC_READ
    else raise EFileMapping.Create('Invalid access mode for file mapping object');

    FFileHandle := CreateFile(PChar(fn), access, 0, sec, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if FFileHandle = INVALID_HANDLE_VALUE then
      raise EFileMapping.CreateFmt('Could not open file %s for file mapping.', [fn]);

    if (size > 0) or (hisize > 0) then begin
      FSize := size;
      FHiSize := hisize;
    end else FSize := GetFileSize(FFileHandle, @FHiSize);
  end;

  try
    FMapHandle := CreateFileMapping(FFileHandle, sec, protect, FHiSize, FSize, PChar(name));
    if FMapHandle = 0 then
      raise EFileMapping.Create('Could not create file mapping object');
  except
    if FFileHandle <> INVALID_HANDLE_VALUE then CloseHandle(FFileHandle);
    raise;
  end;
end;

destructor TFileMapping.Destroy;
begin
  if FMapHandle <> 0 then
    CloseHandle(FMapHandle);

  if FFileHandle <> INVALID_HANDLE_VALUE then
    CloseHandle(FFileHandle);

  inherited;
end;

function TFileMapping.MakeView(mtx: TMutex; start: cardinal; starthi: cardinal = 0; size: cardinal = 0; access: cardinal = 0): TFileView;
// Public function to create a view into a mapped file. Mutex must be used when
// the file mapping object is shared.
begin
  if start mod FPageSize > 0 then
    raise EFileMapping.CreateFmt('Invalid start address %8x for mapped file view', [start]);
  if start + starthi + size > FSize + FHiSize then
    raise EFileMapping.Create('Invalid size for mapped file view');
  if FShared and (mtx = nil) then
    raise EFileMapping.Create('Mutex required for view on shared file mapping object');

  if access = 0 then access := 8 div (FProtect and $E); // inherit access rights

  Result := TFileView.Create(self, mtx, access, size, start, starthi);
end;

//////////////////////////////////////////////////////////////////////////////
// TFileView
constructor TFileView.Create(map: TFileMapping; mtx: TMutex; access, size, start: cardinal; starthi: cardinal = 0);
begin
  FFileMap := map;
  FMutex := mtx;

  if size = 0 then begin
    if map.FHiSize > 0 then
      raise EFileMapping.Create('Mapping object too large for complete view.');
    FSize := map.FSize;
  end else
    FSize := size;

  FMemory := MapViewOfFile(map.FMapHandle, access, starthi, start, size);
  if FMemory = nil then
    raise EFileMapping.Create('View of file mapping object could not be created.');

  FCurrent := FMemory;
  FPosition := 0;
end;

destructor TFileView.Destroy;
begin
  if FMemory <> nil then
    UnmapViewOfFile(FMemory);
end;

procedure TFileView.SetPosition(pos: cardinal);
begin
  if pos >= FSize then
    raise EFileMapping.Create('Invalid position on file view object');

  FPosition := pos;
  FCurrent := pointer(cardinal(FMemory) + pos);
end;

function TFileView.Read(buffer: pointer; count: cardinal): cardinal;
begin
  Result := 0;
  if FMutex.Get then begin
    if FPosition + count > FSize then Result := FSize - FPosition
    else Result := count;

    Move(FCurrent^, buffer^, Result);

    Inc(FPosition, Result);
    Inc(cardinal(FCurrent), Result);

    FMutex.Release;
  end;
end;

function TFileView.Write(buffer: pointer; count: cardinal): cardinal;
begin
  Result := 0;
  if FMutex.Get then begin
    if FPosition + count > FSize then Result := FSize - FPosition
    else Result := count;

    Move(buffer^, FCurrent^, Result);

    Inc(FPosition, Result);
    Inc(cardinal(FCurrent), Result);

    FMutex.Release;
  end;
end;

function TFileView.WriteSafe(buffer: pointer; count: cardinal): cardinal;
begin
  Result := 0;
  if FMutex.Get then begin
    if FPosition + count > FSize then Result := FSize - FPosition
    else Result := count;

    Move(buffer^, FCurrent^, Result);

    try
      if not FlushViewOfFile(FCurrent, Result) then
        raise EFileMapping.CreateFmt('Unable to flush file view at position %d (file address %8x)', [FPosition, FCurrent]);
    finally
      Inc(FPosition, Result);
      Inc(cardinal(FCurrent), Result);

      FMutex.Release;
    end;
  end;
end;

function TFileView.Flush: boolean;
begin
  Result := FMutex.Get;
  if Result then begin
    try
      Result := FlushViewOfFile(FMemory, 0);
      if not Result then
        raise EFileMapping.Create('Unable to flush file view');
    finally
      FMutex.Release;
    end;
  end;
end;

function TFileView.Lock: boolean;
begin
  Result := FMutex.Get;
end;

procedure TFileView.Unlock;
begin
  FMutex.Release;
end;

end.
