//==============================================================================
//    PBTools.pas - CGI Framework module. Implements TDBTagInterpreter
//                  abstract component.
//    Copyright (C) 2002-2003  Ivan Camilo da Cruz
//==============================================================================
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//==============================================================================
//    Este programa  software livre; voc pode redistribu-lo e/ou
//    modific-lo sob os termos da Licena Pblica Geral GNU, conforme
//    publicada pela Free Software Foundation; tanto a verso 2 da
//    Licena como (a seu critrio) qualquer verso mais nova.
//
//    Este programa  distribudo na expectativa de ser til, mas SEM
//    QUALQUER GARANTIA; sem mesmo a garantia implcita de
//    COMERCIALIZAO ou de ADEQUAO A QUALQUER PROPSITO EM
//    PARTICULAR. Consulte a Licena Pblica Geral GNU para obter mais
//    detalhes.
//
//    Voc deve ter recebido uma cpia da Licena Pblica Geral GNU
//    junto com este programa; se no, escreva para a Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
//    02111-1307, USA.
//==============================================================================

unit PBDb;

interface

uses SysUtils, SysConst, Classes, Contnrs, DB, PageBuilder;

type TPBDbRepository = class;
     TDBTagInterpreter = class;

     TPBDbRepositoryItem = class ( TObject )
         private
             fName: String;
             fDataSet: TDataSet;
             fDataSource: TDataSource;
             fTagInterpreter: TDBTagInterpreter;
         public
             property Name: String read fName write fName;
             property DataSet: TDataSet read fDataSet write fDataSet;
             property DataSource: TDataSource read fDataSource write fDataSource;
             property TagInterpreter: TDBTagInterpreter read fTagInterpreter write fTagInterpreter;
             constructor Create ( const pName, pSql: String; const pTagInterpreter: TDBTagInterpreter; const pMaster: TDataSource ); dynamic;
             destructor Destroy; override;
     end;

     TPBDbRepository = class ( TObjectList )
         private
             fTagInterpreter: TDBTagInterpreter;
             function getItem ( index: Integer ): TPBDbRepositoryItem;
             function getItemByName ( index: String ): TPBDbRepositoryItem;
             procedure setItem ( index: Integer; pItem: TPBDbRepositoryItem );
         protected
             function DoCreateRepositoryItem ( const pName, pSql: String; const pMaster: TDataSource ): TPBDbRepositoryItem; dynamic; abstract;
         public
             constructor Create ( const pTagInterpreter: TDBTagInterpreter );
             function Add ( const pName, pMaster, pSql: String ): TPBDbRepositoryItem; overload;
             property Items [ index: Integer ]: TPBDbRepositoryItem read getItem write setItem; default;
             property ItemsByName [ index: String ]: TPBDbRepositoryItem read getItemByName;
             property TagInterpreter: TDBTagInterpreter read fTagInterpreter write fTagInterpreter;
     end;

     TDBTagInterpreter = class ( TCustomTagInterpreter )
         private
             fRepository: TPBDbRepository;
         protected
             fMasterDataSource: TDataSource;
             function DoCreateRepository: TPBDbRepository; dynamic; abstract;
             function DoCreateDataset ( const pSql: String; const pMaster: TDataSource ): TDataSet; dynamic; abstract;
             procedure DoTagEvent ( pageBuilder: TBasePageBuilder; const TagString: string; TagParams: TStrings; const TagContent: string; var ReplaceText: string ); override;
         public
             constructor Create ( pOwner: TComponent ); override;
             destructor Destroy; override;
             procedure Clear; override;
             property Repository: TPBDbRepository read fRepository write fRepository;
         published
     end;

implementation

{ TPBDbRepositoryItem }

constructor TPBDbRepositoryItem.Create ( const pName, pSql: String; const pTagInterpreter: TDBTagInterpreter; const pMaster: TDataSource );
  begin
    inherited Create;

    fName := pName;
    fTagInterpreter := pTagInterpreter;

    fDataSet := pTagInterpreter.DoCreateDataset ( pSql, pMaster );
    fDataSet.Open;

    fDataSource := TDataSource.Create ( nil );
    fDataSource.DataSet := fDataSet;
  end;

destructor TPBDbRepositoryItem.Destroy;
  begin
    fDataset.Free;
    fDataSource.Free;

    inherited;
  end;

{ TPBDatasetList }

constructor TPBDbRepository.Create ( const pTagInterpreter: TDBTagInterpreter );
  begin
    fTagInterpreter := pTagInterpreter;
  end;

function TPBDbRepository.getItem ( index: Integer ): TPBDbRepositoryItem;
  begin
    Result := TPBDbRepositoryItem ( inherited getItem ( index ) );
  end;

function TPBDbRepository.getItemByName ( index: String ): TPBDbRepositoryItem;
  var i: Integer;
  begin
    i := 0;
    while ( i < Count ) and ( CompareText ( Items[i].fName, index ) <> 0 ) do
        Inc ( i );

    if i < Count then
        Result := Items[i]
    else
        raise Exception.Create ( 'Undefined repository item: ' + index );
  end;

procedure TPBDbRepository.setItem ( index: Integer; pItem: TPBDbRepositoryItem );
  begin
    inherited setItem ( index, pItem );
  end;

function TPBDbRepository.Add ( const pName, pMaster, pSql: String ): TPBDbRepositoryItem;
  var masterItem: TPBDbRepositoryItem;
      master: TDataSource;
  begin
    if pMaster <> '' then
      begin
        masterItem := ItemsByName [ pMaster ];
        if masterItem <> nil then
            master := masterItem.fDataSource
        else
            master := fTagInterpreter.fMasterDataSource
      end
    else
        master := fTagInterpreter.fMasterDataSource;

    Result := DoCreateRepositoryItem ( pName, pSql, master );
    Add ( Result );
  end;

{ TDBTagInterpreter }

constructor TDBTagInterpreter.Create(pOwner: TComponent);
  begin
    inherited;

    NameSpace := 'db';
    fRepository := DoCreateRepository;
  end;

destructor TDBTagInterpreter.Destroy;
  begin
    fRepository.Free;

    inherited;
  end;

procedure TDBTagInterpreter.Clear;
  begin
    fRepository.Clear;
  end;

procedure TDBTagInterpreter.DoTagEvent ( pageBuilder: TBasePageBuilder; const TagString: string; TagParams: TStrings; const TagContent: string; var replaceText: string );

    function collectSubDatasets ( master: TPBDbRepositoryItem; html: String ): String;
      var salvaMaster: TDataSource;
      begin
        salvaMaster := fMasterDataSource;
        try
            fMasterDataSource := master.DataSource;
            Result := pageBuilder.ContentFromString ( html );
        finally
            fMasterDataSource := salvaMaster;
        end;
      end;

    function browseDataset ( master: TPBDbRepositoryItem; html, delim, lastDelim: String ): String;
      var firstBlock, lastBlock: String;
      begin
        Result := '';
        with master.Dataset do
          begin
            First;

            if not Eof then
              begin
                // At last one record...

                firstBlock := pageBuilder.ContentFromString ( html );
                Next;

                if not Eof then
                  begin
                    // At last two records...

                    lastBlock := pageBuilder.ContentFromString ( html );
                    Next;

                    Result := firstBlock;

                    if not Eof then
                      begin
                        // Three or more records...

                        while not Eof do
                          begin
                            Result := Result  + delim + lastBlock;

                            lastBlock := pageBuilder.ContentFromString ( html );
                            Next;
                          end;
                      end;

                    Result := Result + lastDelim + lastBlock;
                  end
                else
                    // Just one record...
                    Result := firstBlock;
              end
            else
                // No records on dataset.
          end;
      end;

    function listDatasetField ( master: TPBDbRepositoryItem; field, delim, lastDelim: String ): String;
      var firstBlock, lastBlock: String;
      begin
        Result := '';
        with master.Dataset do
          begin
            First;

            if not Eof then
              begin
                // At last one record...

                firstBlock := FieldByName ( field ).AsString;
                Next;

                if not Eof then
                  begin
                    // At last two records...

                    lastBlock := FieldByName ( field ).AsString;
                    Next;

                    Result := firstBlock;

                    if not Eof then
                      begin
                        // Three or more records...

                        while not Eof do
                          begin
                            Result := Result  + delim + lastBlock;

                            lastBlock := FieldByName ( field ).AsString;
                            Next;
                          end;
                      end;

                    Result := Result + lastDelim + lastBlock;
                  end
                else
                    // Just one record...
                    Result := firstBlock;
              end
            else
                // No records on dataset.
          end;
      end;

    function evaluateIf ( const op1, op2, opr: String ): Boolean;

        function getOpValue ( op: String ): Variant;

          var datasetName, atribute: String;
              p: Integer;
              dataSet : TDataSet;

          begin
            p := Pos ( '.', op );

            if p <> 0 then
              begin
                datasetName := Copy ( op, 1, p - 1 );
                atribute := Copy ( op, p + 1, MaxInt );

                dataSet := Repository.getItemByName ( datasetName ).DataSet;

                if atribute [ 1 ] = '$' then
                  begin
                    atribute := Copy ( atribute, 2, MaxInt );

                    if CompareText ( atribute, 'Eof' ) = 0 then
                        Result := dataSet.Eof
                    else if CompareText ( atribute, 'Bof' ) = 0 then
                        Result := dataSet.Bof
                    else if CompareText ( atribute, 'Empty' ) = 0 then
                        Result := dataSet.IsEmpty
                    else if CompareText ( atribute, 'RecNo' ) = 0 then
                        Result := dataSet.RecNo
                    else if CompareText ( atribute, 'RecCount' ) = 0 then
                        Result := dataSet.RecordCount
                  end
                else
                    Result := dataSet.FieldValues [ atribute ];
              end
            else
                Result := op;
          end;

      var v1, v2: Variant;

      begin
        if op1 = '' then
            if opr = '' then
                // op2 must be boolean.

                Result := getOpValue ( op2 )
            else
              begin
                // opr must be unary.

                if CompareText ( opr, 'not' ) = 0 then
                    Result := not getOpValue ( op2 )
                else
                    raise Exception.Create ( 'Unary operator expected: "' + opr + '" found.' );
              end
        else if op2 = '' then
            if opr = '' then
                // op1 must be boolean.

                Result := getOpValue ( op1 )
            else
                raise Exception.Create ( 'Unary operators must precede operand.' )
        else if opr <> '' then
          begin
            // opr must be binary.

            v1 := getOpValue ( op1 );
            v2 := getOpValue ( op2 );

            if CompareText ( opr, 'eq' ) = 0 then
                Result := v1 = v2
            else if CompareText ( opr, 'ne' ) = 0 then
                Result := v1 <> v2
            else if CompareText ( opr, 'gt' ) = 0 then
                Result := v1 > v2
            else if CompareText ( opr, 'lt' ) = 0 then
                Result := v1 < v2
            else if CompareText ( opr, 'ge' ) = 0 then
                Result := v1 >= v2
            else if CompareText ( opr, 'le' ) = 0 then
                Result := v1 <= v2
            else
                raise Exception.Create ( 'Unary operator expected: "' + opr + '" found.' );
          end
        else
            raise Exception.Create ( 'Operator missed.' )
      end;

  var reposItem: TPBDbRepositoryItem;

  begin
    if CompareText ( tagString, 'db:dataset' ) = 0 then
      begin
        reposItem := Repository.Add ( tagParams.Values [ 'name' ], tagParams.Values [ 'master' ], tagParams.Values [ 'sql' ] );

        if tagContent <> '' then
            replaceText := collectSubDatasets ( reposItem, tagContent )
        else
            replaceText := '';
      end
    else if CompareText ( tagString, 'db:browse' ) = 0 then
      begin
        reposItem := Repository.ItemsByName [ tagParams.Values [ 'dataset' ] ];
        replaceText := browseDataset ( reposItem, tagContent, tagParams.Values [ 'delim' ], tagParams.Values [ 'lastDelim' ] );
      end
    else if CompareText ( tagString, 'db:list' ) = 0 then
      begin
        reposItem := Repository.ItemsByName [ tagParams.Values [ 'dataset' ] ];
        replaceText := listDatasetField ( reposItem, tagParams.Values [ 'field' ], tagParams.Values [ 'delim' ], tagParams.Values [ 'lastDelim' ] );
      end
    else if CompareText ( tagString, 'db:fieldvalue' ) = 0 then
      begin
        reposItem := Repository.ItemsByName [ tagParams.Values [ 'dataset' ] ];
        replaceText := reposItem.Dataset.FieldByName ( tagParams.Values [ 'field' ] ).AsString;
      end
    else if CompareText ( tagString, 'db:if' ) = 0 then
      begin
        if evaluateIf ( tagParams.Values [ 'op1' ], tagParams.Values [ 'op2' ], tagParams.Values [ 'opr' ] ) then
            replaceText := pageBuilder.ContentFromString ( tagContent )
        else
            replaceText := '';
      end;
  end;

end.
