//==============================================================================
//    PBTools.pas - CGI Framework module. Implements TPBTagInterpreter
//                  component.
//    Copyright (C) 2002-2003  Ivan Camilo da Cruz
//==============================================================================
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//==============================================================================
//    Este programa  software livre; voc pode redistribu-lo e/ou
//    modific-lo sob os termos da Licena Pblica Geral GNU, conforme
//    publicada pela Free Software Foundation; tanto a verso 2 da
//    Licena como (a seu critrio) qualquer verso mais nova.
//
//    Este programa  distribudo na expectativa de ser til, mas SEM
//    QUALQUER GARANTIA; sem mesmo a garantia implcita de
//    COMERCIALIZAO ou de ADEQUAO A QUALQUER PROPSITO EM
//    PARTICULAR. Consulte a Licena Pblica Geral GNU para obter mais
//    detalhes.
//
//    Voc deve ter recebido uma cpia da Licena Pblica Geral GNU
//    junto com este programa; se no, escreva para a Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
//    02111-1307, USA.
//==============================================================================

unit PBTools;

interface

uses SysUtils, Classes, Contnrs, PageBuilder;

type TPBTagInterpreter = class;

     TGetVarEvent = procedure ( const sender: TPBTagInterpreter;
                                const varName: String;
                                var value: String ) of object;

     TGetVarMode = ( gvmAllways, gvmNotStored );

     TPBTagInterpreter = class ( TCustomTagInterpreter )
         private
             // Internal
             ifStack: TStringList;
             // Storage for proprerties
             fIncludeDir: String;
             fVarPool: TStringList;
             // Storage for event handlers
             fGetVarMode: TGetVarMode;
             fOnGetVar: TGetVarEvent;
         protected
             procedure DoTagEvent ( pageBuilder: TBasePageBuilder; const tagString: string; tagParams: TStrings; const tagContent: string; var replaceText: string ); override;
         public
             constructor Create ( pOwner: TComponent ); override;
             destructor Destroy; override;
             procedure Clear; override;
         published
             property IncludeDir: String read fIncludeDir write fIncludeDir;
             property VarPool: TStringList read fVarPool write fVarPool;
             property GetVarMode: TGetVarMode read fGetVarMode write fGetVarMode;
             property OnHTMLTag;
             property OnGetVar: TGetVarEvent read fOnGetVar write fOnGetVar;
     end;

function CatPath ( const path1, path2: String ): String;

procedure Register;

implementation

procedure Register;
  begin
    RegisterComponents ( 'CGI Framework', [ TPBTagInterpreter ] );
  end;

function CatPath ( const path1, path2: String ): String;
{$ifdef WIN32}
  const delim: Char = '\';
{$else}
  const delim: Char = '/';
{$endif}
  begin
    if ( path1 = '' ) or ( path2 = '' ) then
        Result := path1 + path2
    else if path1 [ Length ( path1 ) ] = delim then
        if path2 [ 1 ] = delim then
            Result := path1 + Copy ( path2, 2, MaxInt )
        else
            Result := path1 + path2
    else if path2 [ 1 ] = delim then
        Result := path1 + path2
    else
        Result := path1 + delim + path2;
  end;

constructor TPBTagInterpreter.Create ( pOwner: TComponent );
  begin
    inherited;

    NameSpace := 'pb';
    ifStack := TStringList.Create;
    fVarPool := TStringList.Create;
    fGetVarMode := gvmNotStored;
  end;

destructor TPBTagInterpreter.Destroy;
  begin
    fVarPool.Free;
    ifStack.Free;

    inherited;
  end;

procedure TPBTagInterpreter.Clear;
  begin
    fVarPool.Clear;
    ifStack.Clear;
  end;

procedure TPBTagInterpreter.DoTagEvent ( pageBuilder: TBasePageBuilder; const tagString: string; tagParams: TStrings; const tagContent: string; var replaceText: string );

  var templateDir: String;

    procedure includeFile ( fileName: String );
      var incFile: TFileStream;
          i: Integer;
      begin
        incFile := TFileStream.Create ( CatPath ( templateDir, fileName ), fmOpenRead + fmShareDenyWrite );
        try
            // Parameters are included in the begining of fVarPool. This way
            // they take precedence over variables with the same name
            // stored previously.
            //
            // Note: TagParams[0] is the file name and is NOT stored.

            for i := 1 to tagParams.Count - 1 do
                fVarPool.Insert ( 0, tagParams[i] );

            replaceText := pageBuilder.ContentFromStream ( incFile );

            // After file inclusion, remove parameters.

            for i := 1 to tagParams.Count - 1 do
                fVarPool.Delete ( 0 );
        finally
            incFile.Free;
        end;
      end;

    procedure encapsule ( fileName, capsuleContent: String );
      var incFile: TFileStream;
          i: Integer;
      begin
        incFile := TFileStream.Create ( CatPath ( templateDir, fileName ), fmOpenRead + fmShareDenyWrite );
        try
            // Parameters are included in the begining of fVarPool. This way
            // they take precedence over variables with the same name
            // stored previously.
            //
            // Note: TagParams[0] is the file name and is NOT stored.

            for i := 1 to tagParams.Count - 1 do
                fVarPool.Insert ( 0, tagParams[i] );

            // Capsule content goes on a special variable called 'capsuleContent'.

            fVarPool.Insert ( 0, 'capsuleContent=' + pageBuilder.ContentFromString ( capsuleContent ) );

            replaceText := pageBuilder.ContentFromStream ( incFile );

            // After file inclusion, remove parameters and 'capsuleContent' pseudo-var.

            for i := 1 to tagParams.Count do
                fVarPool.Delete ( 0 );
        finally
            incFile.Free;
        end;
      end;

    function getVarValue ( varName: String ): String;
      begin
        if fGetVarMode = gvmAllways then
            if Assigned ( fOnGetVar ) then
                fOnGetVar ( Self, varName, Result )
            else
                Result := fVarPool.Values [ varName ]
        else if ( fGetVarMode = gvmNotStored ) and
                ( fVarPool.IndexOfName ( varName ) = -1 ) then
            if Assigned ( fOnGetVar ) then
                fOnGetVar ( Self, varName, Result )
            else
                Result := ''
        else
            Result := fVarPool.Values [ varName ];
      end;

    procedure manageVars;
      var idxName, idxValue: Integer;
      begin
        idxName  := tagParams.IndexOfName ( 'name' );
        idxValue := tagParams.IndexOfName ( 'value' );

        if idxName <> - 1 then
            // Long sintax...
            if idxValue = -1 then
                // Var replacement, long sintax...
                replaceText := getVarValue ( tagParams.Values [ 'name' ] )
            else
                // Var storage, long sintax...
                fVarPool.Insert ( 0, tagParams.Values [ 'name' ] + '=' + tagParams.Values [ 'value' ] )
        else if idxValue = -1 then
            // Short sintax...

            if tagParams.Count < 1 then
                raise Exception.Create ( 'Too few parameters.' )
            else if tagParams.Count > 1 then
                raise Exception.Create ( 'Too many parameters.' )
            else if AnsiPos ( '=', tagParams [ 0 ] ) = 0 then
                // Var replacement, short sintax...
                replaceText := getVarValue ( tagParams [ 0 ] )
            else
                // Var storage, short sintax...
                fVarPool.Insert ( 0, tagParams [ 0 ] )
        else
            raise Exception.Create ( 'Name is mandatory when using Value parameter.' );
      end;

    procedure deleteVar ( varName: String );
      var i: Integer;
      begin
        i := fVarPool.IndexOfName ( varName );

        if i <> -1 then
            fVarPool.Delete ( i );
      end;

    function evaluateIf ( params: TStrings ): String;

      type TIfItemType = ( iiEof, iiNot, iiEq, iiNe, iiGt, iiGe, iiLt, iiLe, iiValue );

      const ifOperators: Set of TIfItemType = [ iiNot..iiLe ];

      var itemIndex: Integer;
          itemType: TIfItemType;
          itemValue: String;

        procedure getNextItem;
          var p: Integer;
              item: String;
          begin
            if itemIndex = params.Count then
              begin
                itemType := iiEof;
                itemValue := '(end of tag)';
              end
            else
              begin
                item := params [ itemIndex ];

                p := AnsiPos ( '=', item );
                if p = 0 then
                    if CompareText ( item, 'not' ) = 0 then
                      begin
                        itemType := iiNot;
                        itemValue := item;
                      end
                    else if CompareText ( item, 'eq' ) = 0 then
                      begin
                        itemType := iiEq;
                        itemValue := item;
                      end
                    else if CompareText ( item, 'ne' ) = 0 then
                      begin
                        itemType := iiNe;
                        itemValue := item;
                      end
                    else if CompareText ( item, 'gt' ) = 0 then
                      begin
                        itemType := iiGt;
                        itemValue := item;
                      end
                    else if CompareText ( item, 'lt' ) = 0 then
                      begin
                        itemType := iiLt;
                        itemValue := item;
                      end
                    else if CompareText ( item, 'ge' ) = 0 then
                      begin
                        itemType := iiGe;
                        itemValue := item;
                      end
                    else if CompareText ( item, 'le' ) = 0 then
                      begin
                        itemType := iiLe;
                        itemValue := item;
                      end
                    else
                      begin
                        itemType := iiValue;
                        itemValue := getVarValue ( item );
                      end
                else if CompareText ( Copy ( item, 1, p - 1 ), 'value' ) = 0 then
                  begin
                    itemType := iiValue;
                    itemValue := Copy ( item, p + 1, MaxInt );
                  end
                else if CompareText ( Copy ( item, 1, p - 1 ), 'var' ) = 0 then
                  begin
                    itemType := iiValue;
                    itemValue := getVarValue ( Copy ( item, p + 1, MaxInt ) );
                  end;
              end;

            Inc ( itemIndex );
          end;

      var resComp: Integer;
          op1: String;
          opr: TIfItemType;

      begin
        itemIndex := 0;
        getNextItem;

        if itemType = iiEof then
            raise Exception.Create ( 'No parameters given for pb:if' )
        else if itemType = iiValue then
          begin
            op1 := itemValue;

            getNextItem;

            if itemType = iiEof then
                Result := op1
            else if itemType in ifOperators then
              begin
                opr := itemType;

                getNextItem;
                if itemType = iiValue then
                  begin
                    resComp := CompareText ( op1, itemValue );

                    if opr = iiEq then
                        if resComp = 0 then
                            Result := '*'
                        else
                            Result := ''
                    else if opr = iiNe then
                        if resComp <> 0 then
                            Result := '*'
                        else
                            Result := ''
                    else if opr = iiGt then
                        if resComp > 0 then
                            Result := '*'
                        else
                            Result := ''
                    else if opr = iiLt then
                        if resComp < 0 then
                            Result := '*'
                        else
                            Result := ''
                    else if opr = iiGe then
                        if resComp >= 0 then
                            Result := '*'
                        else
                            Result := ''
                    else if opr = iiLe then
                        if resComp <= 0 then
                            Result := '*'
                        else
                            Result := ''
                  end
                else
                    raise Exception.Create ( 'A value was expected after binary operator in pb:if' )
              end
            else
                raise Exception.Create ( 'Binary operator expected: "' + itemValue + '" found in pb:if.' );
          end
        else if itemType = iiNot then
          begin
            getNextItem;

            if itemType = iiValue then
                if itemValue = '' then
                    Result := '*'
                else
                    Result := ''
            else
                raise Exception.Create ( 'A value was expected after a not in pb:if' )
          end
        else
            raise Exception.Create ( 'Value or unary operator expected: "' + itemValue + '" found in pb:if.' );
      end;

  begin
    // not inherited

    if IncludeDir <> '' then
        templateDir := IncludeDir
    else if TCustomPageBuilder ( pageBuilder ).HTMLFile <> '' then   // Ugly, no?
        templateDir := ExtractFilePath ( TCustomPageBuilder ( pageBuilder ).HTMLFile )
    else
        templateDir := GetCurrentDir;

    if CompareText ( tagString, 'pb:include' ) = 0 then
        includeFile ( tagParams.Values [ 'filename' ] )
    else if CompareText ( tagString, 'pb:includeindirect' ) = 0 then
        includeFile ( getVarValue ( tagParams.Values [ 'varname' ] ) )
    else if CompareText ( tagString, 'pb:encapsule' ) = 0 then
        encapsule ( tagParams.Values [ 'filename' ], tagContent )
    else if CompareText ( tagString, 'pb:encapsuleindirect' ) = 0 then
        encapsule ( getVarValue ( tagParams.Values [ 'varname' ] ), tagContent )
    else if CompareText ( tagString, 'pb:var' ) = 0 then
        manageVars
    else if CompareText ( tagString, 'pb:clearVar' ) = 0 then
        deleteVar ( tagParams.Values [ 'name' ] )
    else if CompareText ( tagString, 'pb:pageTitle' ) = 0 then
        replaceText := fVarPool.Values [ 'pageTitle' ]
    else if CompareText ( tagString, 'pb:capsuleContent' ) = 0 then
        replaceText := fVarPool.Values [ 'capsuleContent' ]
    else if CompareText ( tagString, 'pb:if' ) = 0 then
      begin
        ifStack.Add ( evaluateIf ( tagParams ) );
        replaceText := pageBuilder.ContentFromString ( tagContent );
        ifStack.Delete ( ifStack.Count - 1 );
      end
    else if CompareText ( tagString, 'pb:then' ) = 0 then
        if ifStack.Count = 0 then
            raise Exception.Create ( 'pb:then without a pb:if' )
        else if ifStack [ ifStack.Count - 1 ] <> '' then
            replaceText := pageBuilder.ContentFromString ( tagContent )
        else
            replaceText := ''
    else if CompareText ( tagString, 'pb:else' ) = 0 then
        if ifStack.Count = 0 then
            raise Exception.Create ( 'pb:else without a pb:if' )
        else if ifStack [ ifStack.Count - 1 ] = '' then
            replaceText := pageBuilder.ContentFromString ( tagContent )
        else
            replaceText := ''
  end;

end.

