//==============================================================================
//    PageBuilder.pas - CGI Framework module. Implements TPageBuilder and
//                      TTagInterpreter components.
//    Copyright (C) 2002-2003  Ivan Camilo da Cruz
//==============================================================================
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//==============================================================================
//    Este programa  software livre; voc pode redistribu-lo e/ou
//    modific-lo sob os termos da Licena Pblica Geral GNU, conforme
//    publicada pela Free Software Foundation; tanto a verso 2 da
//    Licena como (a seu critrio) qualquer verso mais nova.
//
//    Este programa  distribudo na expectativa de ser til, mas SEM
//    QUALQUER GARANTIA; sem mesmo a garantia implcita de
//    COMERCIALIZAO ou de ADEQUAO A QUALQUER PROPSITO EM
//    PARTICULAR. Consulte a Licena Pblica Geral GNU para obter mais
//    detalhes.
//
//    Voc deve ter recebido uma cpia da Licena Pblica Geral GNU
//    junto com este programa; se no, escreva para a Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
//    02111-1307, USA.
//==============================================================================

unit PageBuilder;

interface

uses SysUtils, Classes;

type TBasePageBuilder = class;

     TPageBuilderTagEvent = procedure ( sender: TBasePageBuilder;
                                        const tagString: string;
                                        tagParams: TStrings;
                                        const tagContent: string;
                                        var replaceText: string ) of object;

     TCustomTagInterpreter = class ( TComponent )
         private
             fOnHTMLTag: TPageBuilderTagEvent;
             fNameSpace: String;
         protected
             procedure DoTagEvent ( pageBuilder: TBasePageBuilder; const TagString: string; TagParams: TStrings; const TagContent: string; var ReplaceText: string ); virtual;
             procedure Clear; virtual; abstract;
         public
             property NameSpace: String read fNameSpace write fNameSpace;
             property OnHTMLTag: TPageBuilderTagEvent read fOnHTMLTag write fOnHTMLTag;
     end;

     TTagInterpreter = class ( TCustomTagInterpreter )
         published
             property NameSpace;
             property OnHTMLTag;
     end;

     TTagInterpreterItem = class ( TCollectionItem )
         private
             fTagInterpreter: TCustomTagInterpreter;
             fActive: Boolean;
             fName: String;
         protected
             function getDisplayName: String; override;
             procedure setDisplayName ( const Value: String ); override;
         published
             constructor Create ( Collection: TCollection ); override;
             property Active: Boolean read fActive write fActive;
             property TagInterpreter: TCustomTagInterpreter read fTagInterpreter write fTagInterpreter;
             property Name: String read fName write setDisplayName;
     end;

     TTagInterpreterCollection = class ( TCollection )
         private
             fOwner: TBasePageBuilder;
             function getTagInterpreter ( Index: Integer ): TTagInterpreterItem;
             procedure setTagInterpreter ( Index: Integer; Value: TTagInterpreterItem );
         protected
             function getAttrCount: Integer; override;
             function getAttr ( Index: Integer): string; override;
             function getItemAttr ( Index, ItemIndex: Integer): string; override;
             function getOwner: TPersistent; override;
             procedure setItemName ( item: TCollectionItem ); override;
         public
            constructor Create ( AOwner: TBasePageBuilder; ItemClass: TCollectionItemClass );
            function Add: TTagInterpreterItem;
            property Items [ Index: Integer ]: TTagInterpreterItem read getTagInterpreter write setTagInterpreter; default;
     end;

     TBasePageBuilder = class ( TComponent )
         private
             fOnHTMLTag: TPageBuilderTagEvent;
             fOnlyRegiteredNamespaces: Boolean;
             fParseAttributes: Boolean;
             fNameSpace: String;
             fTagInterpreters: TTagInterpreterCollection;
             function isTagWithNameSpace ( tag: String ): Boolean;
             function isTagWithRegisteredNameSpace ( tag: String ): Boolean;
             function getNamespace ( tag: String ): String;
             function getTagInterpreter ( namespace: String ): TCustomTagInterpreter;
         protected
             procedure DoTagEvent ( const TagString: string; TagParams: TStrings; const TagContent: string; var ReplaceText: string ); dynamic;
             function GetTemplateFileName: string; virtual;
             function GetTemplateStream ( out AOwned: Boolean ): TStream; virtual;
             property OnHTMLTag: TPageBuilderTagEvent read fOnHTMLTag write fOnHTMLTag;
         public
             constructor Create ( AOwner: TComponent ); override;
             destructor Destroy; override;
             function Content: String;
             function ContentFromStream ( Stream: TStream ): string;
             function ContentFromString ( const s: string ): string;
             property ParseAttributes: Boolean read fParseAttributes write fParseAttributes default true;
             property OnlyRegiteredNamespaces: Boolean read fOnlyRegiteredNamespaces write fOnlyRegiteredNamespaces default false;
             property NameSpace: String read fNameSpace write fNameSpace;
             property TagInterpreters: TTagInterpreterCollection read fTagInterpreters write fTagInterpreters;
     end;

     TCustomPageBuilder = class ( TBasePageBuilder )
         private
             fHTMLFile: String;
             fHTMLDoc: TStrings;
             procedure SetHTMLFile ( const Value: String );
             procedure SetHTMLDoc ( Value: TStrings );
         protected
             function GetTemplateStream ( out AOwned: Boolean ): TStream; override;
         public
             constructor Create(AOwner: TComponent); override;
             destructor Destroy; override;
             property HTMLDoc: TStrings read fHTMLDoc write SetHTMLDoc;
             property HTMLFile: String read fHTMLFile write SetHTMLFile;
     end;

     TPageBuilder = class ( TCustomPageBuilder )
         published
             property HTMLDoc;
             property HTMLFile;
             property ParseAttributes;
             property OnlyRegiteredNamespaces;
             property NameSpace;
             property TagInterpreters;
             property OnHTMLTag;
     end;

procedure Register;

implementation

procedure Register;
  begin
    RegisterComponents ( 'CGI Framework', [ TTagInterpreter, TPageBuilder ] );
  end;

type TToken = ( tkEof, tkNonParsedBlock, tkTag, tkStartTag, tkEndTag );

     THtmlParser = class ( TObject )
         private
             fInput: TStream;
             fToken: TToken;
             fChar: Char;
             fNonParsedBlock: String;
             fTag: String;
             fTagParams: TStringList;
             procedure firstChar;
             procedure nextChar;
         public
             constructor Create ( pInput: TStream );
             destructor Destroy; override;
             property Token: TToken read FToken;
             property NonParsedBlock: String read fNonParsedBlock;
             property Tag: String read fTag;
             property TagParams: TStringList read fTagParams;
             procedure FirstToken;
             procedure NextToken;
     end;

const firstTagIdChar = [ 'a'..'z', 'A'..'Z', '0'..'9' ];
      nextTagIdChar = firstTagIdChar + [ ':', '-', '_' ];
      lineEndChar = [ #10, #12, #13 ];
      spacingChar = [ ' ', #9 ] + lineEndChar;

{ THtmlParser }

constructor THtmlParser.Create ( pInput: TStream );
  begin
    inherited Create;

    fInput := pInput;
    fToken := tkNonParsedBlock;
    fNonParsedBlock := '';
    fTagParams := TStringList.Create;
  end;

destructor THtmlParser.Destroy;
  begin
    fTagParams.Free;

    inherited;
  end;

procedure THtmlParser.firstChar;
  begin
    fInput.Seek ( 0, soFromBeginning );
    nextChar;
  end;

procedure THtmlParser.nextChar;
  begin
    fNonParsedBlock := fNonParsedBlock + fChar;
    if fInput.Read ( fChar, 1 ) = 0 then
        fChar := #0;
  end;

procedure THtmlParser.FirstToken;
  begin
    firstChar;
    if fToken <> tkEof then
        NextToken
  end;

procedure THtmlParser.NextToken;

    procedure skipSpaces;
      begin
        while fChar in spacingChar do
            nextChar;
      end;

    procedure getTagId;
      begin
        fTag := '';
        while fChar in nextTagIdChar do
          begin
            fTag := fTag + fChar;
            nextChar;
          end;
      end;

    procedure getTagParams;

        procedure getOneTagParam;
          var paramId, paramValue: String;
              savePreviousChar: Char;
          begin
            paramId := '';
            while fChar in nextTagIdChar do
              begin
                paramId := paramId + fChar;
                nextChar;
              end;

            skipSpaces;

            if fChar = '=' then
              begin
                nextChar;
                skipSpaces;

                paramValue := '';

                if ( fChar = '"' ) or ( fChar = '''' ) then
                  begin
                    // Attibute value will be one or more blocks of strings
                    // delimited by single (') or double (") quotes.

                    while ( fChar = '"' ) or ( fChar = '''' ) do
                      begin
                        // Begin block of delimited attribute value...

                        if fChar = '"' then
                          begin
                            // That block is delimited by double quote...

                            nextChar;
                            while not ( fChar in lineEndChar ) and
                                 ( fChar <> '"' ) and
                                 ( fChar <> #0 ) do
                              begin
                                if fChar = '''' then
                                  begin
                                    savePreviousChar := fChar;
                                    nextChar;
                                    if fChar = '''' then
                                      begin
                                        paramValue := paramValue + '"';
                                        nextChar;
                                      end
                                    else
                                        paramValue := paramValue + savePreviousChar;
                                  end
                                else
                                  begin
                                    paramValue := paramValue + fChar;
                                    nextChar;
                                  end
                              end;

                            if fChar = '"' then
                                nextChar;
                          end
                        else if fChar = '''' then
                          begin
                            // And this one, by single quotes...

                            nextChar;
                            while not ( fChar in lineEndChar ) and
                                  ( fChar <> '''' ) and
                                  ( fChar <> #0 ) do
                              begin
                                if fChar = '''' then
                                  begin
                                    savePreviousChar := fChar;
                                    nextChar;
                                    if fChar = '''' then
                                      begin
                                        paramValue := paramValue + '''';
                                        nextChar;
                                      end
                                    else
                                        paramValue := paramValue + savePreviousChar;
                                  end
                                else
                                  begin
                                    paramValue := paramValue + fChar;
                                    nextChar;
                                  end
                              end;

                            if fChar = '''' then
                                nextChar;
                          end;

                        // After the delimited string, looks for the next block...

                        skipSpaces;
                      end
                  end
                else
                  begin
                    // Attribute value is delimited by the first spacing character...

                    while not ( fChar in spacingChar ) and not ( fChar in [ #0, '/', '>' ] ) do
                      begin
                        paramValue := paramValue + fChar;
                        nextChar;
                      end;
                  end;

                fTagParams.Add ( paramId + '=' + paramValue );
              end
            else
              begin
                // Tag attribute without a value.

                fTagParams.Add ( paramId );
              end;
          end;

      begin   // getTagParams
        while fChar in firstTagIdChar do
          begin
            getOneTagParam;
            skipSpaces;
          end;
      end;

  begin
    fNonParsedBlock := '';
    fTagParams.Clear;
    if fChar = '<' then
      begin
        nextChar;
        if fChar in firstTagIdChar then
          begin
            getTagId;

            // Temos o Id da tag. Analiza parmetros (se existirem).

            skipSpaces;

            if fChar in firstTagIdChar then
                getTagParams;

            if fChar = '/' then
              begin
                nextChar;

                if fChar = '>' then
                  begin
                    nextChar;
                    fToken := tkTag;
                  end
                else
                    fToken := tkNonParsedBlock;
              end
            else if fChar = '>' then
              begin
                nextChar;
                fToken := tkStartTag;
              end;
          end
        else if fChar = '/' then
          begin
            nextChar;
            getTagId;
            if fChar = '>' then
              begin
                nextChar;
                fToken := tkEndTag;
              end
            else
                fToken := tkNonParsedBlock;
          end
        else
            fToken := tkNonParsedBlock;
      end
    else if fChar <> #0 then
      begin
        while ( fChar <> '<' ) and ( fChar <> #0 ) do
            nextChar;

        fToken := tkNonParsedBlock;
      end
    else
        fToken := tkEof;
  end;

{ TCustomTagInterpreter }

procedure TCustomTagInterpreter.DoTagEvent ( pageBuilder: TBasePageBuilder; const TagString: string; TagParams: TStrings; const TagContent: string; var ReplaceText: string );
  begin
    ReplaceText := '';

    if Assigned ( fOnHtmlTag ) then
        fOnHTMLTag ( pageBuilder, TagString, TagParams, TagContent, ReplaceText );
  end;

{ TTagInterpreterItem }

constructor TTagInterpreterItem.Create ( Collection: TCollection );
  begin
    inherited Create ( Collection );

    fActive := true;
  end;

function TTagInterpreterItem.getDisplayName: String;
  begin
    Result := fName;
  end;

procedure TTagInterpreterItem.setDisplayName ( const Value: String );
  var i: Integer;
      interpreter: TTagInterpreterItem;
  begin
    if AnsiCompareText ( Value, FName ) <> 0 then
      begin
        if Collection <> nil then
            for I := 0 to Collection.Count - 1 do
              begin
                interpreter := TTagInterpreterCollection ( Collection ).Items[I];

                if ( interpreter <> Self) and
                   ( interpreter is TTagInterpreterItem ) and
                   ( AnsiCompareText ( Value, interpreter.Name ) = 0 ) then
                    raise Exception.Create ( 'Existe outro tem de lista de interpretadores com este mesmo nome.' );
              end;

        fName := Value;

        Changed ( False );
      end;
  end;

{ TTagInterpreterCollection }

function TTagInterpreterCollection.getTagInterpreter ( Index: Integer ): TTagInterpreterItem;
  begin
    Result := TTagInterpreterItem ( inherited Items [ index ] );
  end;

procedure TTagInterpreterCollection.setTagInterpreter ( Index: Integer; Value: TTagInterpreterItem );
  begin
    Items [ index ].Assign ( value );
  end;

function TTagInterpreterCollection.getAttrCount: Integer;
  begin
    Result := 3;
  end;

function TTagInterpreterCollection.getAttr ( Index: Integer ): string;
  begin
    case index of
        0: Result := 'TagInterpreter';
        1: Result := 'NameSpace';
        2: Result := 'State';
    end;
  end;

function TTagInterpreterCollection.getItemAttr ( Index, ItemIndex: Integer ): string;
  begin
    if TTagInterpreterItem ( Items [ ItemIndex ] ).TagInterpreter <> nil then
        case index of
            0: Result := TTagInterpreterItem ( Items [ ItemIndex ] ).TagInterpreter.Name;
            1: Result := TTagInterpreterItem ( Items [ ItemIndex ] ).TagInterpreter.NameSpace;
            2: if TTagInterpreterItem ( Items [ ItemIndex ] ).Active then
                   Result := 'Active'
               else
                   Result := 'Inactive';
        end
    else
        Result := 'nil';
  end;

function TTagInterpreterCollection.Add: TTagInterpreterItem;
  begin
    Result := TTagInterpreterItem ( inherited Add );
  end;

constructor TTagInterpreterCollection.Create ( AOwner: TBasePageBuilder; ItemClass: TCollectionItemClass );
  begin
    inherited Create ( ItemClass );

    fOwner := AOwner;
  end;

function TTagInterpreterCollection.getOwner: TPersistent;
  begin
    Result := fOwner;
  end;

procedure TTagInterpreterCollection.setItemName ( item: TCollectionItem );
  var i, tenta: Integer;
      itemName: string;
      curItem: TTagInterpreterItem;
      found: Boolean;
  begin
    tenta := 0;

    repeat
        Inc ( tenta );
        itemName := fOwner.Name + Format ( 'TagInterpreter%d', [ tenta ] );

        found := false;
        i := 0;
        while ( i < Count ) and ( not found ) do
          begin
            curItem := TTagInterpreterItem ( Items[i] );

            found := ( curItem <> item ) and
                     ( CompareText ( curItem.Name, itemName ) = 0 );

            Inc ( i );
          end;
    until not found;

    TTagInterpreterItem ( Item ).Name := itemName;
  end;

{ TBasePageBuilder }

function TBasePageBuilder.getNamespace(tag: String): String;
  begin
    Result := Copy ( tag, 1, Pos ( ':', tag ) - 1 );
  end;

function TBasePageBuilder.isTagWithNameSpace ( tag: String ): Boolean;
  begin
    Result := Pos ( ':', tag ) <> 0;
  end;

function TBasePageBuilder.isTagWithRegisteredNameSpace ( tag: String ): Boolean;
  var namespace: String;
  begin
    namespace := getNamespace ( tag );

    if namespace <> '' then
        Result := ( CompareText ( fNameSpace, namespace ) = 0 ) or
                  ( getTagInterpreter ( namespace ) <> nil )
    else
        Result := false;
  end;

function TBasePageBuilder.getTagInterpreter ( namespace: String ): TCustomTagInterpreter;
  var i: Integer;
  begin
    if namespace <> '' then
      begin
        i := 0;
        while ( i < fTagInterpreters.Count ) and
              ( CompareText ( fTagInterpreters[i].fTagInterpreter.NameSpace, namespace ) <> 0 ) do
            Inc ( i );

        if i < fTagInterpreters.Count then
            Result := fTagInterpreters[i].TagInterpreter
        else
            Result := nil;
      end
    else
        Result := nil;
  end;

function TBasePageBuilder.Content: string;
  var entrada: TStream;
      owned: Boolean;
  begin
    Result := '';
    entrada := GetTemplateStream ( Owned );
    try
        if entrada <> nil then
            Result := ContentFromStream ( entrada );
    finally
        if owned then
            entrada.Free;
    end;
  end;

function TBasePageBuilder.ContentFromStream ( stream: TStream ): string;

  var parser: THtmlParser;

    function parseAttributes: Boolean;
      var i, p: Integer;
          attValue: String;
      begin
        Result := false;

        if fParseAttributes then
          begin
            for i := 0 to parser.TagParams.Count - 1 do
              begin
                p := AnsiPos ( '=', parser.TagParams [ i ] );
                if p <> 0 then
                  begin
                    attValue := Copy ( parser.TagParams [ i ], p + 1, MaxInt );

                    if Pos ( '<', attValue ) <> 0 then
                      begin
                        parser.TagParams [ i ] := Copy ( parser.TagParams [ i ], 1, p ) + ContentFromString ( attValue );
                        Result := true;
                      end;
                  end;
              end;
          end;
      end;

    function mustWeReplaceThatTag ( tag: String ): Boolean;
      begin
        if fOnlyRegiteredNamespaces then
            Result := isTagWithRegisteredNameSpace ( tag )
        else
            Result := isTagWithNameSpace ( tag );
      end;

    function rebuildTag: String;
      var i, p: Integer;
      begin
        Result := '<' + parser.Tag;
        for i := 0 to parser.TagParams.Count - 1 do
          begin
            p := Pos ( '=', parser.TagParams[i] );
            if p = 0 then
                Result := Result + ' ' + parser.TagParams[i]
            else if Pos ( '"', parser.TagParams[i] ) = 0 then
                Result := Result + ' ' + Copy ( parser.TagParams[i], 1, p ) + '"' + Copy ( parser.TagParams[i], p + 1, MaxInt ) + '"'
            else
                Result := Result + ' ' + Copy ( parser.TagParams[i], 1, p ) + '''' + Copy ( parser.TagParams[i], p + 1, MaxInt ) + '''';
          end;

        Result := Result + '>';
      end;

  var output: TStringStream;
      tagName, tagContent, replaceText: String;
      tagParams: TStringList;
      level: Integer;

  begin
    output := TStringStream.Create ( '' );
    parser := THtmlParser.Create ( stream );
    try
        parser.FirstToken;
        while parser.Token <> tkEof do
          begin
            if parser.Token = tkNonParsedBlock then
                output.WriteString ( parser.NonParsedBlock )
            else if Assigned ( fOnHtmlTag ) or ( fTagInterpreters.Count > 0 ) then
              begin
                if parser.Token = tkTag then
                  begin
                    // We have a simple tag...

                    if mustWeReplaceThatTag ( parser.Tag ) then
                      begin
                        parseAttributes;

                        DoTagEvent ( parser.Tag, parser.TagParams, '', replaceText );
                        output.WriteString ( replaceText );
                      end
                    else if parseAttributes then
                        output.WriteString ( rebuildTag )
                    else
                        output.WriteString ( parser.NonParsedBlock );
                  end
                else if parser.Token = tkStartTag then
                  begin
                    // We have a start-tag.

                    if mustWeReplaceThatTag ( parser.Tag ) then
                      begin
                        parseAttributes;

                        tagName := parser.Tag;
                        tagParams := TStringList.Create;
                        tagParams.AddStrings ( parser.TagParams );
                        try
                            tagContent := '';
                            level := 1;

                            repeat
                                parser.NextToken;

                                if ( parser.Token = tkStartTag ) and ( CompareText ( parser.Tag, tagName ) = 0 ) then
                                    Inc ( level )
                                else if ( parser.Token = tkEndTag ) and ( CompareText ( parser.Tag, tagName ) = 0 ) then
                                    Dec ( level );

                                if level <> 0 then
                                    tagContent := tagContent + parser.NonParsedBlock;

                            until ( parser.Token = tkEof ) or ( level = 0 );

                            DoTagEvent ( tagName, tagParams, tagContent, replaceText );

                            output.WriteString ( replaceText );
                        finally
                            tagParams.Free;
                        end;
                      end
                    else if parseAttributes then
                        output.WriteString ( rebuildTag )
                    else
                        output.WriteString ( parser.NonParsedBlock );
                  end
                else
                    // End-tag without start tag? Put on output...
                    output.WriteString ( parser.NonParsedBlock );
              end
            else
              begin
                if isTagWithNameSpace ( parser.Tag ) then
                  begin
                    DoTagEvent ( parser.Tag, parser.TagParams, tagContent, replaceText );
                    output.WriteString ( replaceText );
                  end
                else
                    output.WriteString ( parser.NonParsedBlock );
              end;

            parser.NextToken;
          end;

        Result := output.DataString;
    finally
        output.Free;
    end;
  end;

function TBasePageBuilder.ContentFromString ( const s: string ): string;
  var entrada: TStringStream;
  begin
    entrada := TStringStream.Create ( s );
    try
        Result := ContentFromStream ( entrada );
    finally
        entrada.Free;
    end;
  end;

constructor TBasePageBuilder.Create ( AOwner: TComponent );
  begin
    inherited;

    fOnlyRegiteredNamespaces := false;
    fParseAttributes := true;
    fTagInterpreters := TTagInterpreterCollection.Create ( Self, TTagInterpreterItem );
  end;

destructor TBasePageBuilder.Destroy;
  begin
    fTagInterpreters.Free;

    inherited;
  end;

procedure TBasePageBuilder.DoTagEvent ( const TagString: string;
                                        TagParams: TStrings;
                                        const TagContent: string;
                                        var ReplaceText: string );
  var tagInterpreter: TCustomTagInterpreter;
  begin
    ReplaceText := '';
    tagInterpreter := getTagInterpreter ( getNamespace ( TagString ) );

    if tagInterpreter <> nil then
        tagInterpreter.DoTagEvent ( Self, TagString, TagParams, TagContent, ReplaceText )
    else if Assigned ( fOnHtmlTag ) then
        fOnHTMLTag ( Self, TagString, TagParams, TagContent, ReplaceText );
  end;

function TBasePageBuilder.GetTemplateFileName: string;
  begin
    Result := '';
  end;

function TBasePageBuilder.GetTemplateStream ( out AOwned: Boolean ): TStream;
  begin
    Result := nil;
  end;

{ TCustomPageBuilder }

constructor TCustomPageBuilder.Create ( AOwner: TComponent );
  begin
    inherited;
    fHTMLDoc := TStringList.Create;
  end;

destructor TCustomPageBuilder.Destroy;
  begin
    fHTMLDoc.Free;
    inherited;
  end;

function TCustomPageBuilder.GetTemplateStream ( out AOwned: Boolean ): TStream;
  begin
    if fHTMLFile <> '' then
      begin
        Result := TFileStream.Create ( fHTMLFile, fmOpenRead + fmShareDenyWrite );
        AOwned := True;
      end
    else
        Result := TStringStream.Create ( FHTMLDoc.Text );
  end;

procedure TCustomPageBuilder.SetHTMLDoc ( Value: TStrings );
  begin
    fHTMLFile := '';
    fHTMLDoc.Assign ( value );
  end;

procedure TCustomPageBuilder.SetHTMLFile ( const Value: String );
  begin
    fHTMLDoc.Clear;
    fHTMLFile := Value;
  end;

end.
